import os
import sys
import traci

import sumo_gym.typing
from sumo_gym.utils.fmp_utils import IDLE_LOCATION


STOPPED_STATUS = 1


class SumoRender:
    def __init__(
        self,
        sumo_gui_path: str = None,
        sumo_config_path: str = None,
        edge_dict: dict = None,
        edge_length_dict: dict = None,
        ev_dict: dict = None,
        edges: sumo_gym.typing.EdgeType = None,
        n_vehicle: int = 1,
    ):
        self.sumo_gui_path = sumo_gui_path
        self.edge_dict = edge_dict
        self.edge_length_dict = edge_length_dict
        self.ev_dict = ev_dict
        self.edges = edges
        self.initialized = False
        self.terminated = False
        self.stop_statuses = [False] * n_vehicle
        self.n_vehicle = n_vehicle
        self.routes = []

        if "SUMO_HOME" in os.environ:
            tools = os.path.join(os.environ["SUMO_HOME"], "tools")
            sys.path.append(tools)
        else:
            sys.exit("please declare environment variable 'SUMO_HOME'")

        traci.start([self.sumo_gui_path, "-c", sumo_config_path])

    def get_stop_status(self):
        return self.stop_statuses

    def update_travel_vertex_info_for_vehicle(self, vehicle_travel_info_list):
        self.travel_info = vehicle_travel_info_list

    def render(self):

        # while not self.terminated:
        if not self.initialized:
            print("Initialize the first starting edge for vehicle...")
            self.initialized = True
            self._initialize_route()
            self._park_vehicle_to_assigned_starting_pos()
        else:
            self._update_route_with_stop()
            traci.simulationStep()
            self._update_stop_status()

    def close(self):
        while not self.terminated:
            traci.simulationStep()

            self.terminated = True

            eligible_vehicle = traci.vehicle.getIDList()
            for i in range(self.n_vehicle):
                vehicle_id = self._find_key_from_value(self.ev_dict, i)
                if vehicle_id not in eligible_vehicle:
                    continue

                if (
                    traci.vehicle.getStopState(vehicle_id) != STOPPED_STATUS
                ):  # as long as one vehicle not arrive its assigned last vertex, continue simulation
                    self.terminated = False

        traci.close()

    def _initialize_route(self):
        for i in range(self.n_vehicle):

            vehicle_id = self._find_key_from_value(self.ev_dict, i)
            edge_id = traci.vehicle.getRoute(vehicle_id)[0]

            # stop at the ending vertex of vehicle's starting edge
            # notice here each vehicle must finish traveling along it starting edge
            # there is no way to reassign it.
            self.routes.append(tuple([edge_id]))

            traci.vehicle.setStop(
                vehID=vehicle_id,
                edgeID=edge_id,
                pos=self.edge_length_dict[edge_id],
                laneIndex=0,
                duration=189999999999,
                flags=0,
                startPos=0,
            )

            print("Step stop for vehicle: ", vehicle_id)

    def _park_vehicle_to_assigned_starting_pos(self):
        print("Parking all car to their destinations of the starting edges...")
        while False in self.stop_statuses:
            traci.simulationStep()
            self._update_stop_status()
        print("All vehicles ready for model routing.")

    def _update_route_with_stop(self):
        """
        Update the route for each vehicle with the edge from its current stopped vertex to the next assigned vertex,
        and set the next stop to that 'to' vertex.
        Skip the vehicles that are still traveling along the edge, i.e., ev_stop = False.
        """
        eligible_vehicle = traci.vehicle.getIDList()
        for i in range(self.n_vehicle):
            if self.stop_statuses[i]:

                # handle the case when the destination of last demand is the start of current demand
                # i.e., the vehicle need to perform drop and pickup at the same location for two different demands correspondingly
                # so the location won't change, stop remains the same, no re-routing needed.
                if self.travel_info[i][0] == self.travel_info[i][1]:
                    continue

                self.stop_statuses[i] = False
                vehicle_id = self._find_key_from_value(self.ev_dict, i)
                if vehicle_id not in eligible_vehicle:
                    continue

                # all demands satisfied or being responding, remove all idle car to unblock others at junction
                if self.travel_info[i][1] == IDLE_LOCATION:
                    traci.vehicle.remove(vehicle_id)
                    print("Vehicle ", vehicle_id, " becomes idle, remove from network.")

                else:
                    traci.vehicle.resume(vehicle_id)

                    via_edge = self.travel_info[i]
                    edge_id = self._find_key_from_value(
                        self.edge_dict, self._find_edge_index(via_edge)
                    )

                    if "split" not in edge_id:
                        actual_edge_id = edge_id
                    else:
                        actual_edge_id = edge_id[7:]
                    self.routes[i] += tuple([actual_edge_id])

                    print(
                        "Vehicle ",
                        vehicle_id,
                        " has arrived stopped location, reassign new routes: ",
                        self.routes[i][-2],
                        self.routes[i][-1],
                    )
                    if (
                        self.routes[i][-1] != self.routes[i][-2]
                    ):  # handle the case for stopping at CS and then resume
                        traci.vehicle.setRoute(
                            vehID=vehicle_id, edgeList=self.routes[i][-2:]
                        )
                    else:
                        edge_id = actual_edge_id
                    traci.vehicle.setStop(
                        vehID=vehicle_id,
                        edgeID=actual_edge_id,
                        pos=self.edge_length_dict[edge_id],
                        laneIndex=0,
                        duration=189999999999,
                        flags=0,
                        startPos=0,
                    )

    def _update_stop_status(self):
        eligible_vehicle = traci.vehicle.getIDList()

        for i in range(self.n_vehicle):
            vehicle_id = self._find_key_from_value(self.ev_dict, i)
            if vehicle_id not in eligible_vehicle:
                continue
            if (
                traci.vehicle.getStopState(vehicle_id) == STOPPED_STATUS
            ):  # arrived the assigned vertex, can be assigned to the next
                self.stop_statuses[i] = True

    def _find_key_from_value(self, dict, value):
        return list(dict.keys())[list(dict.values()).index(value)]

    def _find_edge_index(self, via_edge):
        for i in range(len(self.edges)):
            if via_edge[0] == self.edges[i].start and via_edge[1] == self.edges[i].end:
                return i



import numpy as np
import numpy.typing as npt
from typing import Dict
import sumo_gym.typing


def calculate_dist(i, j, vertices) -> float:
    l1, l2 = vertices[i], vertices[j]
    try:
        return np.sqrt(np.power(l1.x - l2.x, 2) + np.power(l1.y - l2.y, 2))
    except:
        return np.sqrt(np.power(l1[0] - l2[0], 2) + np.power(l1[1] - l2[1], 2))


def get_adj_to_list(vertices, edges) -> sumo_gym.typing.AdjListType:
    try:
        adj = [[] for _ in range(len(vertices))]
        for e in edges:
            adj[e.start].append(e.end)
        return np.asarray(adj, dtype=object)
    except:
        adj = [[] for _ in range(len(vertices))]
        for e in edges:
            adj[e[0]].append(e[1])

        return np.asarray(adj, dtype=object)


def get_adj_from_list(vertices, edges) -> sumo_gym.typing.AdjListType:
    try:
        adj = [[] for _ in range(len(vertices))]
        for e in edges:
            adj[e.end].append(e.start)
        return np.asarray(adj, dtype=object)
    except:
        adj = [[] for _ in range(len(vertices))]
        for e in edges:
            adj[e[1]].append(e[0])

        return np.asarray(adj, dtype=object)



import sumo_gym.utils.network_utils as network_utils
import numpy as np
import numpy.typing as npt
from typing import Tuple

NO_LOADING = -1
NO_CHARGING = -1
CHARGING_STATION_LENGTH = 5
IDLE_LOCATION = -1


class Vertex(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, other):
        return (self.x, self.y) == (other.x, other.y)

    def __lt__(self, other):
        return (self.x, self.y) < (other.x, other.y)

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return f"vertex ({self.x}, {self.y})"


class Edge(object):
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __eq__(self, other):
        return (self.start, self.end) == (other.start, other.end)

    def __lt__(self, other):
        return (self.start, self.end) < (other.start, other.end)

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return f"edge ({self.start}, {self.end})"


class Demand(object):
    def __init__(self, departure, destination):
        self.departure = departure
        self.destination = destination

    def __eq__(self, other):
        return (self.departure, self.destination) == (
            other.departure,
            other.destination,
        )

    def __lt__(self, other):
        return (self.departure, self.destination) < (other.departure, other.destination)

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return f"demand ({self.departure}, {self.destination})"


class ElectricVehicles(object):
    def __init__(self, id, speed, indicator, capacity):
        self.id = id
        self.speed = speed
        self.indicator = indicator
        self.capacity = capacity

    def __eq__(self, other):
        return self.location == other.location

    def __lt__(self, other):
        return self.location < other.location

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return f"ElectricVehicles ({self.id}, {self.speed}, {self.indicator}, {self.capacity})"


class ChargingStation(object):
    def __init__(self, location, indicator, charging_speed):
        self.location = location
        self.indicator = indicator
        self.charging_speed = charging_speed

    def __eq__(self, other):
        return self.location == other.location

    def __lt__(self, other):
        return self.location < other.location

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return f"ChargingStation ({self.location}, {self.indicator}, {self.charging_speed})"


class Loading(object):
    def __init__(self, current=-1, target=-1):
        self.current = current
        self.target = target

    def __repr__(self):
        return f"(responding {self.current}, goto respond {self.target})"


class Charging(object):
    def __init__(self, current=-1, target=-1):
        self.current = current
        self.target = target

    def __repr__(self):
        return f"(charging {self.current}, go to charge {self.target})"


class GridAction(object):
    def __init__(self, state=None):
        self.is_loading = state.is_loading
        self.is_charging = state.is_charging
        self.location = state.location

    def __repr__(self):
        return f"({self.is_loading}, {self.is_charging}, location {self.location})"


class FMPState(object):
    def __init__(
        self,
        location=0,
        is_loading=Loading(-1, -1),
        is_charging=Charging(-1, -1),
        battery=0,
    ):
        self.location = location
        self.is_loading = is_loading
        self.is_charging = is_charging
        self.battery = battery
        self.stopped = False  # arrive the assigned vertex

    def __repr__(self):
        return (
            f"Location: {self.location}, "
            + f"Is loading: {(self.is_loading.current, self.is_loading.target)},"
            + f"Is charging: {(self.is_charging)} "
            + f"Battery: {(self.battery)}"
        )


def convert_raw_vertices(raw_vertices):
    """
    Each raw vertex is [id (str), x_coord (float), y_coord (float)]
    """
    vertices = []
    vertex_dict = {}  # vertex id in SUMO to idx in vertices
    for counter, v in enumerate(raw_vertices):
        vertices.append(Vertex(v[1], v[2]))
        vertex_dict[v[0]] = counter
    return vertices, vertex_dict


def convert_raw_edges(raw_edges, vertex_dict):
    """
    Each raw edge is
    [id (str), from_vertex_id (str), to_vertex_id (str), edge_length (float)]
    """
    edges = []
    edge_dict = {}  # sumo edge_id to idx in edges
    edge_length_dict = {}  # sumo edge_id to length
    for counter, e in enumerate(raw_edges):
        new_edge = Edge(vertex_dict[e[1]], vertex_dict[e[2]])
        edges.append(new_edge)
        edge_dict[e[0]] = counter
        edge_length_dict[e[0]] = e[3]
    return edges, edge_dict, edge_length_dict


def euclidean_distance(start_x, start_y, end_x, end_y):
    """
    Compute euclidean distance between (start_x, start_y)
    and (end_x, end_y)
    """
    return (((start_x - end_x) ** 2) + ((start_y - end_y) ** 2)) ** 0.5


def convert_raw_charging_stations(
    raw_charging_stations, vertices, edges, edge_dict, edge_length_dict
):
    """
    Each raw charging station is
    [id, (x_coord, y_coord), edge_id, charging speed]
    """

    charging_station_dict = {}  # idx in charging_stations to sumo id
    charging_stations = []

    vtx_counter = len(vertices)

    for counter, charging_station in enumerate(raw_charging_stations):

        charging_station_dict[counter] = charging_station[0]

        # create new vertex with charging station's location
        x_coord, y_coord = charging_station[1]
        new_vtx = Vertex(x_coord, y_coord)
        vertices.append(new_vtx)

        # create two new edges
        # first get the start and end vertex indices of the old edge
        edge_id = charging_station[2]
        old_edge_start_idx = edges[edge_dict[edge_id]].start
        old_edge_end_idx = edges[edge_dict[edge_id]].end

        edge_length_positive_edge_cs = charging_station[4]
        edge_length_postive_edge = edge_length_dict[edge_id]

        curr_edge_count = len(edge_dict)
        edges.append(Edge(old_edge_start_idx, vtx_counter))
        edge_dict["split1_%s" % edge_id] = curr_edge_count
        edge_length_dict["split1_%s" % edge_id] = edge_length_positive_edge_cs

        curr_edge_count += 1
        edges.append(Edge(vtx_counter, old_edge_start_idx))
        edge_dict["split1_-%s" % edge_id] = curr_edge_count
        edge_length_dict["split1_-%s" % edge_id] = edge_length_positive_edge_cs

        curr_edge_count += 1
        edges.append(Edge(vtx_counter, old_edge_end_idx))
        edge_dict["split2_%s" % edge_id] = curr_edge_count
        edge_length_dict["split2_%s" % edge_id] = (
            edge_length_postive_edge
            - edge_length_positive_edge_cs
            + CHARGING_STATION_LENGTH
        )

        curr_edge_count += 1
        edges.append(Edge(old_edge_end_idx, vtx_counter))
        edge_dict["split2_-%s" % edge_id] = curr_edge_count
        edge_length_dict["split2_-%s" % edge_id] = (
            edge_length_postive_edge
            - edge_length_positive_edge_cs
            + CHARGING_STATION_LENGTH
        )

        # instantiate new ChargingStation with location set to idx in `vertices`
        charging_stations.append(ChargingStation(vtx_counter, 220, charging_station[3]))

        vtx_counter += 1

    return charging_stations, charging_station_dict, edge_length_dict


def convert_raw_electric_vehicles(raw_electric_vehicles):
    """
    Each raw electric vehicle is
    [id (str), maximum speed (float), maximumBatteryCapacity (float)]
    """

    electric_vehicles = []
    ev_dict = {}  # ev sumo id to idx in electric_vehicles
    for counter, vehicle in enumerate(raw_electric_vehicles):
        electric_vehicles.append(
            ElectricVehicles(vehicle[0], vehicle[1], 220, vehicle[2])
        )

        ev_dict[vehicle[0]] = counter

    return electric_vehicles, ev_dict


def convert_raw_departures(raw_departures, ev_dict, edges, edge_dict, num_vehicles):
    """
    Each raw departure is [vehicle_id, starting_edge_id]
    """
    departures = np.zeros(num_vehicles)
    actual_departures = np.zeros(num_vehicles)
    for dpt in raw_departures:
        actual_departures[ev_dict[dpt[0]]] = edges[edge_dict[dpt[1]]].start
        departures[ev_dict[dpt[0]]] = edges[edge_dict[dpt[1]]].end
    return departures, actual_departures


def convert_raw_demand(raw_demand, vertex_dict):
    """
    Each raw demand is [junction_id, dest_vertex_id]
    """
    demand = []
    for d in raw_demand:
        demand.append(Demand(vertex_dict[d[0]], vertex_dict[d[1]]))
    return demand


def one_step_to_destination(vertices, edges, start_index, dest_index):
    if start_index == dest_index:
        return dest_index
    visited = [False] * len(vertices)
    bfs_queue = [dest_index]
    visited[dest_index] = True

    while bfs_queue:
        curr = bfs_queue.pop(0)
        adjacent_map = network_utils.get_adj_from_list(vertices, edges)

        for v in adjacent_map[curr]:
            if not visited[v] and v == start_index:
                return curr
            elif not visited[v]:
                bfs_queue.append(v)
                visited[v] = False


def nearest_charging_station_with_distance(
    vertices, charging_stations, edges, start_index
):
    charging_station_vertices = [
        charging_station.location for charging_station in charging_stations
    ]
    visited = [False] * len(vertices)

    bfs_queue = [[start_index, 0]]
    visited[start_index] = True

    while bfs_queue:
        curr, curr_depth = bfs_queue.pop(0)
        adjacent_map = network_utils.get_adj_to_list(vertices, edges)

        for v in adjacent_map[curr]:
            if not visited[v] and v in charging_station_vertices:
                return charging_station_vertices.index(v), curr_depth + 1
            elif not visited[v]:
                bfs_queue.append([v, curr_depth + 1])
                visited[v] = False


def dist_between(vertices, edges, start_index, dest_index):
    if start_index == dest_index:
        return 0
    visited = [False] * len(vertices)
    bfs_queue = [[start_index, 0]]
    visited[start_index] = True
    while bfs_queue:
        curr, curr_depth = bfs_queue.pop(0)
        adjacent_map = network_utils.get_adj_to_list(vertices, edges)

        for v in adjacent_map[curr]:
            if not visited[v] and v == dest_index:
                return curr_depth + 1
            elif not visited[v]:
                bfs_queue.append([v, curr_depth + 1])
                visited[v] = False


def get_hot_spot_weight(vertices, edges, demands, demand_start):
    adjacent_vertices = np.append(
        network_utils.get_adj_to_list(vertices, edges)[demand_start], demand_start
    )
    local_demands = len([d for d in demands if d.departure in adjacent_vertices])

    return local_demands / len(demands) * 100



import xml.etree.ElementTree as ET
import numpy as np
from typing import Dict, Any, List, Tuple
import numpy.typing as npt
import sys

VEHICLE_XML_TAG = "vehicle"
VEHICLE_CAPACITY_TAG = "personNumber"

VERTEX_XML_TAG = "junction"
VERTEX_XML_INVALID_TYPE = "internal"
VERTEX_CUSTOMIZED_PARAM = "param"
VERTEX_DEMAND_KEY = "destination"

EDGE_XML_TAG = "edge"
EDGE_XML_INVALID_FUNC = "internal"
EDGE_XML_PRIORITY = "-1"


def get_electric_vehicles(flow_xml_tree):
    """
    Helper function for decode_xml_fmp

    Returns electric vehicles

    Each EV is [id (str), maximum speed (float), maximumBatteryCapacity (float)]
    """

    # there should only be one vType
    # defined in rou.xml

    # in this model, all EV's start off fully charged
    vtype = flow_xml_tree.findall("vType")[0]
    vtype_params = vtype.findall("param")
    maxBatteryCap = -1
    for vtype_param in vtype_params:
        if vtype_param.attrib["key"] == "maximumBatteryCapacity":
            maxBatteryCap = float(vtype_param.attrib["value"])

    ev_lst = []
    vehicles = flow_xml_tree.findall(VEHICLE_XML_TAG)
    for vehicle in vehicles:
        ev_lst.append(
            [vehicle.attrib["id"], float(vtype.attrib["maxSpeed"]), maxBatteryCap]
        )
    return ev_lst


def get_charging_stations(additional_xml_tree, net_xml_tree):
    """
    Helper function for decode_xml_fmp

    Returns charging stations

    Each charging station is [id, (x_coord, y_coord), edge_id, charging speed]
    """
    cs_lst = []
    stations = additional_xml_tree.findall("chargingStation")
    for station in stations:

        if "shadow" not in station.attrib["id"]:

            # get approximate location
            x_coord, y_coord = station.findall("param")[0].attrib["value"].split()
            # get edge_id
            lane_id = station.attrib["lane"]
            edge_id = None
            # get all edges in net.xml
            edges = net_xml_tree.findall("edge")
            for edge in edges:
                lanes = edge.findall("lane")
                lanes = [
                    lane.attrib["id"] for lane in lanes if lane.attrib["id"] == lane_id
                ]
                if len(lanes) == 1:
                    edge_id = edge.attrib["id"]
                    break
            cs_lst.append(
                (
                    station.attrib["id"],
                    (float(x_coord), float(y_coord)),
                    edge_id,
                    float(station.attrib["power"]),
                    float(station.attrib["endPos"]),
                )
            )

    return cs_lst


def get_vertices(net_xml_tree):
    """
    Helper function for decode_xml_fmp

    Returns vertices
    Each vertex is [id (str), x_coord (float), y_coord (float)]
    """
    vtx_lst = []
    junctions = net_xml_tree.findall(VERTEX_XML_TAG)

    for junction in junctions:
        if junction.attrib["type"] == VERTEX_XML_INVALID_TYPE:
            continue
        vtx_lst.append(
            [
                junction.attrib["id"],
                float(junction.attrib["x"]),
                float(junction.attrib["y"]),
            ]
        )

    return vtx_lst


def get_edges(net_xml_tree):
    """
    Helper function for decode_xml_fmp

    Returns edges

    Each edge is [id (str), from_vertex_id (str),
                  to_vertex_id (str),
                  edge_length (float)]
    """
    edge_lst = []
    edges = net_xml_tree.findall(EDGE_XML_TAG)
    for e in edges:
        if "function" in e.attrib and e.attrib["function"] == EDGE_XML_INVALID_FUNC:
            continue
        # Edge lengths are given by lane lengths. Each edge has at
        # least one lane and all lanes of an edge have
        # the same length
        lane = e.findall("lane")[0]
        edge_lst.append(
            [
                e.attrib["id"],
                e.attrib["from"],
                e.attrib["to"],
                float(lane.attrib["length"]),
            ]
        )
    return edge_lst


def get_departures(flow_xml_tree):
    """
    Helper function for decode_xml_fmp

    Returns departures for each vehicle

    Each departure is [vehicle_id, starting_edge_id]
    and should be defined for all vehicles
    """
    departures = []  # id, start_index tuple
    for vehicle in flow_xml_tree.findall(VEHICLE_XML_TAG):
        route = vehicle.findall("route")[0]  # findall should return one
        edges = route.attrib["edges"]  # space separated list of edge ids
        start_edge = edges.split()[0]
        departures.append([vehicle.attrib["id"], start_edge])

    return departures


def get_demand(net_xml_tree):
    """
    Some junctions are customer nodes and this is represented by
    a junction having param destination
    Each demand is [junction_id, dest_vertex_id]
    """
    demand = []
    for junction in net_xml_tree.findall(VERTEX_XML_TAG):
        if junction.get("type") != VERTEX_XML_INVALID_TYPE:
            for customized_params in junction.findall(VERTEX_CUSTOMIZED_PARAM):
                if customized_params.get("key") == VERTEX_DEMAND_KEY:
                    demand.append([junction.get("id"), customized_params.get("value")])

    return demand


def decode_xml_fmp(
    net_xml_file_path: str = None,
    flow_xml_file_path: str = None,
    additional_xml_path: str = None,
):
    """
    Parse files generated from SUMO and return a FMP instance

    Returns vertices, charging_stations, electric_vehicles,
    edges, departures, and demands
    """
    net_xml_tree = ET.parse(net_xml_file_path)
    flow_xml_tree = ET.parse(flow_xml_file_path)
    additional_xml_tree = ET.parse(additional_xml_path)
    vertices = get_vertices(net_xml_tree)
    charging_stations = get_charging_stations(additional_xml_tree, net_xml_tree)
    electric_vehicles = get_electric_vehicles(flow_xml_tree)
    edges = get_edges(net_xml_tree)
    departures = get_departures(flow_xml_tree)
    demand = get_demand(net_xml_tree)
    return vertices, charging_stations, electric_vehicles, edges, departures, demand


def decode_xml(
    net_xml_file_path: str = None, flow_xml_file_path: str = None
) -> Tuple[npt.NDArray[Any]]:
    """
    Parse the net.xml and rou.xml generated from SUMO and read into VRP initialization environments.
    Return objects: vertices, demand, edges, departures, capacity for VRP class
    """

    net_xml_source = open(net_xml_file_path)
    flow_xml_source = open(flow_xml_file_path)

    vertices, demand, edge_id_map, edges = _parse_network_xml(net_xml_source)
    departures, capacity = _parse_flow_xml(flow_xml_source, edge_id_map, edges)

    net_xml_source.close()
    flow_xml_source.close()

    return (
        np.asarray(vertices),
        np.asarray(demand),
        np.asarray(edges),
        np.asarray(departures),
        np.asarray(capacity),
    )


def _parse_flow_xml(flow_file_path: str, edge_id_map: Dict[str, int], edges: Any):
    """
    :param flow_file_path:      file path of rou.xml
    :param edge_id_map:         sample structure: {'genE0': 0, 'genE1': 1}
    :param edges:               tuple of [edge_from_vertex, edge_to_vertex], sample structure: [[0,4], [1,3], [7,5]]
    """
    flow_tree = ET.parse(flow_file_path)
    flow_xml_root = flow_tree.getroot()

    departures = []  # int array
    capacity = []  # float array
    for vehicle_trip in flow_xml_root.findall(VEHICLE_XML_TAG):
        departure_edge = vehicle_trip.get("from")
        departures.append(edges[edge_id_map[departure_edge]][0])

        capacity_value = vehicle_trip.get(VEHICLE_CAPACITY_TAG) or 20.0
        capacity.append(float(capacity_value))

    return departures, capacity


def _parse_network_xml(network_file_path: str):
    """
    :param network_file_path:     file path of net.xml
    """
    network_tree = ET.parse(network_file_path)
    network_xml_data = network_tree.getroot()

    vertices_id_map = {}  # sample structure: {'genJ1': 0, 'genJ10': 1}
    vertices = []  # tuple of x,y position of each vertex
    demand = []  # float array
    vertex_count = 0
    for junction in network_xml_data.findall(VERTEX_XML_TAG):
        if junction.get("type") != VERTEX_XML_INVALID_TYPE:
            vertices.append([float(junction.get("x")), float(junction.get("y"))])
            vertices_id_map[junction.get("id")] = vertex_count

            demand_value = 0.0
            for customized_params in junction.findall(VERTEX_CUSTOMIZED_PARAM):
                if customized_params.get("key") == VERTEX_DEMAND_KEY:
                    demand_value = float(customized_params.get("value"))
            demand.append(demand_value)

            vertex_count += 1

    edge_id_map = {}  # sample structure: {'genE0': 0, 'genE1': 1}
    edges = (
        []
    )  # tuple of [edge_from_vertex, edge_to_vertex], sample structure: [[0,4], [1,3], [7,5]]
    edge_count = 0
    for edge in network_xml_data.findall(EDGE_XML_TAG):
        if edge.get("priority") == EDGE_XML_PRIORITY:
            edges.append(
                [vertices_id_map[edge.get("from")], vertices_id_map[edge.get("to")]]
            )
            edge_id_map[edge.get("id")] = edge_count
            edge_count += 1

    return vertices, demand, edge_id_map, edges



import sys
import sumo_gym.envs as envs
import numpy as np
from typing import Dict, Any, Union, Set

try:
    import matplotlib
    import matplotlib.pyplot as plt
    import matplotlib.transforms as transforms
except ModuleNotFoundError:
    print(
        "sumo_gym requires matplotlib to plot, either install sumo_gym[plot] or matplotlib",
        file=sys.stderr,
    )
    raise


def _expand_shortcuts(key: str) -> str:
    if key == "ls":
        return "linestyle"
    return key


def _filter_dict(
    __dict: Dict[str, Any], prefix: str, *, ignore: Union[Set[str], None] = None
) -> Dict[str, Any]:
    """
    Keyword argument conversion: convert the kwargs to several independent args, pulling
    them out of the dict given. Prioritize prefix_kw dict.
    """

    # If passed explicitly, use that
    if f"{prefix}kw" in __dict:
        res: dict[str, Any] = __dict.pop(f"{prefix}kw")
        return {_expand_shortcuts(k): v for k, v in res.items()}

    ignore_set: set[str] = ignore or set()
    return {
        _expand_shortcuts(key[len(prefix) :]): __dict.pop(key)
        for key in list(__dict)
        if key.startswith(prefix) and key not in ignore_set
    }


def plot_VRP(self: envs.vrp.VRP, *, ax: matplotlib.axes.Axes, **kwargs) -> Any:
    x, y = list(), list()
    for v in self.vertices:
        x.append(v[0])
        y.append(v[1])

    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(1, 0, 0)

    depot_kwargs = _filter_dict(kwargs, "depot_")
    vertex_kwargs = _filter_dict(kwargs, "vertex_")
    edge_kwargs = _filter_dict(kwargs, "edge_")
    if len(kwargs):
        raise ValueError(f"{set(kwargs)} not needed")

    # avoid alpha=0 when no demand
    if max(self.demand) == 0:
        alpha = np.ones(self.demand.shape) * 0.6
    else:
        alpha = self.demand / max(self.demand) * 0.4 + np.ones(self.demand.shape) * 0.6

    for e in self.edges:
        l1, l2 = self.vertices[e[0]], self.vertices[e[1]]
        ax.plot([l1[0], l2[0]], [l1[1], l2[1]], **edge_kwargs)

    ax.scatter(x[: self.n_depot], y[: self.n_depot], alpha=1, **depot_kwargs)
    return ax.scatter(
        x[self.n_depot :],
        y[self.n_depot :],
        alpha=alpha[self.n_depot :],
        **vertex_kwargs,
    )


def plot_VRPEnv(
    self: envs.vrp.VRPEnv,
    *,
    ax_dict: Union[Dict[str, matplotlib.axes.Axes], None] = None,
    **kwargs: Any,
) -> Any:
    if ax_dict is None:
        ax_dict = dict()

    if ax_dict:
        try:
            vrp_ax = ax_dict["vrp_ax"]
            demand_ax = ax_dict["demand_ax"]
            loading_ax = ax_dict["loading_ax"]
        except KeyError as err:
            raise ValueError("All axes should be all given or none at all") from err
    else:
        fig = plt.gcf()
        grid = fig.add_gridspec(
            2, 2, hspace=0, wspace=0, width_ratios=[4, 1], height_ratios=[1, 4]
        )
        vrp_ax = fig.add_subplot(grid[1, 0])
        demand_ax = fig.add_subplot(grid[0, 0], sharex=vrp_ax)
        loading_ax = fig.add_subplot(grid[1, 1], sharey=vrp_ax)

    # keyword arguments
    vrp_kwargs = _filter_dict(kwargs, "vrp_", ignore={"vrp_alpha"})
    location_kwargs = _filter_dict(kwargs, "location_")
    demand_kwargs = _filter_dict(kwargs, "demand_")
    loading_kwargs = _filter_dict(kwargs, "loading_")
    if len(kwargs):
        raise ValueError(f"{set(kwargs)} not needed")

    vrp_art = plot_VRP(self.vrp, ax=vrp_ax, **vrp_kwargs)
    x = [self.vrp.vertices[l][0] for l in self.locations]
    y = [self.vrp.vertices[l][1] for l in self.locations]
    vrp_ax.scatter(x, y, alpha=1, **location_kwargs)
    demand_art = demand_ax.bar(
        np.arange(self.vrp.n_vertex), self.vrp.demand, **demand_kwargs
    )
    demand_ax.spines["top"].set_visible(False)
    demand_ax.spines["right"].set_visible(False)
    demand_ax.xaxis.set_visible(False)
    demand_ax.set_ylabel("Demand")
    base = loading_ax.transData
    rot = transforms.Affine2D().rotate_deg(90).scale(-1, 1)
    loading_art = loading_ax.bar(
        np.arange(self.vrp.n_vehicle),
        self.loading,
        transform=rot + base,
        **loading_kwargs,
    )
    loading_ax.spines["top"].set_visible(False)
    loading_ax.spines["right"].set_visible(False)
    loading_ax.yaxis.set_visible(False)
    loading_ax.set_xlabel("Loading")

    return vrp_art, demand_art, loading_art


def plot_FMP(self: envs.fmp.FMP, *, ax: matplotlib.axes.Axes, **kwargs) -> Any:
    x, y = list(), list()
    for v in self.vertices:
        x.append(v.x)
        y.append(v.y)

    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(1, 0, 0)

    vertex_kwargs = _filter_dict(kwargs, "vertex_")
    edge_kwargs = _filter_dict(kwargs, "edge_")
    if len(kwargs):
        raise ValueError(f"{set(kwargs)} not needed")

    # avoid alpha=0 when no demand
    # if max(self.demand) == 0:
    #     alpha = np.ones(self.demand.shape) * 0.6
    # else:
    #     alpha = self.demand / max(self.demand) * 0.4 + np.ones(self.demand.shape) * 0.6

    for e in self.edges:
        l1, l2 = self.vertices[e.start], self.vertices[e.end]
        ax.plot([l1.x, l2.x], [l1.y, l2.y], **edge_kwargs)

    # ax.scatter(x[: self.n_depot], y[: self.n_depot], alpha=1, **depot_kwargs)
    return ax.scatter(
        x,
        y,
        **vertex_kwargs,
    )


def plot_FMPEnv(
    self: envs.fmp.FMPEnv,
    *,
    ax_dict: Union[Dict[str, matplotlib.axes.Axes], None] = None,
    **kwargs: Any,
) -> Any:
    if ax_dict is None:
        ax_dict = dict()

    if ax_dict:
        try:
            fmp_ax = ax_dict["fmp_ax"]
            demand_ax = ax_dict["demand_ax"]
            loading_ax = ax_dict["loading_ax"]
        except KeyError as err:
            raise ValueError("All axes should be all given or none at all") from err
    else:
        fig = plt.gcf()
        grid = fig.add_gridspec(
            2, 2, hspace=0, wspace=0, width_ratios=[4, 1], height_ratios=[1, 4]
        )
        fmp_ax = fig.add_subplot(grid[1, 0])
        demand_ax = fig.add_subplot(grid[0, 0], sharex=fmp_ax)
        loading_ax = fig.add_subplot(grid[1, 1], sharey=fmp_ax)

    # keyword arguments
    fmp_kwargs = _filter_dict(kwargs, "fmp_", ignore={"fmp_alpha"})
    location_kwargs = _filter_dict(kwargs, "location_")
    # demand_kwargs = _filter_dict(kwargs, "demand_")
    # loading_kwargs = _filter_dict(kwargs, "loading_")
    # if len(kwargs):
    #     raise ValueError(f"{set(kwargs)} not needed")

    fmp_art = plot_FMP(self.fmp, ax=fmp_ax, **fmp_kwargs)
    x = [self.fmp.vertices[s.location].x for s in self.states]
    y = [self.fmp.vertices[s.location].y for s in self.states]
    fmp_ax.scatter(x, y, alpha=1, **location_kwargs)
    # demand_art = demand_ax.bar(
    #     np.arange(self.fmp.n_vertex), self.fmp.demand, **demand_kwargs
    # )
    demand_ax.spines["top"].set_visible(False)
    demand_ax.spines["right"].set_visible(False)
    demand_ax.xaxis.set_visible(False)
    demand_ax.set_ylabel("Demand")
    # base = loading_ax.transData
    # rot = transforms.Affine2D().rotate_deg(90).scale(-1, 1)
    # loading_art = loading_ax.bar(
    #     np.arange(self.fmp.n_vehicle),
    #     self.loading,
    #     transform=rot + base,
    #     **loading_kwargs,
    # )
    loading_ax.spines["top"].set_visible(False)
    loading_ax.spines["right"].set_visible(False)
    loading_ax.yaxis.set_visible(False)
    loading_ax.set_xlabel("Loading")
    return fmp_art



from typing import Tuple
from sumo_gym.utils.xml_utils import decode_xml
from sumo_gym.utils.network_utils import (
    calculate_dist,
    get_adj_from_list,
    get_adj_to_list,
)
from sumo_gym.utils.svg_uitls import vehicle_marker

__all__ = (
    "decode_xml",
    "calculate_dist",
    "get_adj_from_list",
    "get_adj_to_list",
    "vehicle_marker",
)


def __dir__() -> Tuple[str, ...]:
    return __all__



from typing import Tuple
from sumo_gym.spaces.network import NetworkSpace
from sumo_gym.spaces.grid import GridSpace

__all__ = ("Network", "Grid")


def __dir__() -> Tuple[str, ...]:
    return __all__



from sumo_gym.envs.vrp import VRP, VRPEnv
from sumo_gym.envs.fmp import FMP, FMPEnv
from typing import Tuple


__all__ = ("VRP", "VRPEnv", "FMP", "FMPEnv")


def __dir__() -> Tuple[str, ...]:
    return __all__



import operator
import numpy as np
import random
import numpy.typing as npt
from typing import Type, Tuple, Dict, Any
import sumo_gym.typing

import gym
from gym import error, spaces, utils
from gym.utils import seeding
import sumo_gym
import sumo_gym.utils.network_utils as network_utils
from sumo_gym.utils.svg_uitls import vehicle_marker


class VRP(object):
    def __init__(
        self,
        net_xml_file_path: str = None,
        demand_xml_file_path: str = None,
        n_vertex: int = 0,
        n_depot: int = 1,
        n_edge: int = 0,
        n_vehicle: int = 0,
        vertices: sumo_gym.typing.VerticesType = None,
        demand: sumo_gym.typing.DemandType = None,
        edges: sumo_gym.typing.EdgeType = None,
        departures: sumo_gym.typing.DeparturesType = None,
        capacity: sumo_gym.typing.CapacityType = None,
    ):
        """
        :param n_vertex:      the number of vertices
        :param n_depot:       the number of depots
        :param n_edge:        the number of edges
        :param n_vehicle:     the number of vehicles
        :param vertices:        the vertices
        :param demand:          the demand of vertices
        :param edges:           the edges
        :param departures:      the departures of vehicles
        :param capacity:        the capacity of vehicles
        Create a vehicle routing problem setting (CVRP if capacity is activated).
        """
        if net_xml_file_path is None or demand_xml_file_path is None:
            # number
            self.n_vertex = n_vertex
            self.n_depot = n_depot
            self.n_edge = n_edge
            self.n_vehicle = n_vehicle

            # network
            self.vertices = vertices
            self.depots = np.asarray(range(n_depot)).astype(np.int32)
            self.demand = demand
            self.edges = edges

            # vehicles
            self.departures = departures
            self.capacity = (
                np.asarray([float("inf") for _ in range(n_vehicle)])
                if capacity is None
                else capacity
            )

            if not self._is_valid():
                raise ValueError("VRP setting is not valid")

        else:
            self.n_depot = 1  # default value

            # read in the sumo xml files and parse them into VRP initial problem settings
            (
                self.vertices,
                self.demand,
                self.edges,
                self.departures,
                self.capacity,
            ) = sumo_gym.utils.decode_xml(net_xml_file_path, demand_xml_file_path)
            self.n_vertex, _ = self.vertices.shape
            self.n_edge, _ = self.edges.shape
            self.n_vehicle = self.departures.shape[0]
            self.depots = self.vertices[: self.n_depot]

    def __repr__(self):
        return (
            f"Vehicle routing problem with {self.n_vertex} vertices, {self.n_depot} depots,"
            + f" {self.n_edge} edges, and {self.n_vehicle} vehicles.\nVertices are {self.vertices};\n"
            + f"Depots are {self.depots};\nDemand are {self.demand};\nEdges are {self.edges};\nDepartures are"
            + f" {self.departures};\nCapacity are {self.capacity}.\n"
        )

    def _is_valid(self):
        if (
            not self.n_vertex
            or not self.n_depot
            or not self.n_edge
            or not self.n_vehicle
            or self.vertices.any() is None
            or self.demand.any() is None
            or self.edges.any() is None
            or self.departures.any() is None
            or self.capacity.any() is None
        ):
            return False
        # todo: scale judgement
        return True

    def get_adj_list(self) -> sumo_gym.typing.AdjListType:
        return network_utils.get_adj_to_list(self.vertices, self.edges)


class VRPEnv(gym.Env):
    metadata = {"render.modes": ["human"]}
    vrp = property(operator.attrgetter("_vrp"))
    __isfrozen = False

    def __init__(self, **kwargs):
        self._vrp = VRP(**kwargs)  # todo: make it "final"
        self.run = -1
        self._reset()
        self._freeze()

    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError(
                "Cannot add new attributes once instance %r is initialized" % self
            )
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__isfrozen = True

    def reset(self):
        return self._reset()

    def _reset(self):
        self.run += 1
        self.locations: sumo_gym.typing.LocationsType = self.vrp.departures.astype(int)
        self.loading: sumo_gym.typing.LoadingType = np.zeros(self.vrp.n_vehicle)
        self.action_space: sumo_gym.spaces.network.NetworkSpace = (
            sumo_gym.spaces.network.NetworkSpace(
                self.locations,
                self.vrp.get_adj_list(),
                self.vrp.demand,
                np.asarray([False] * self.vrp.n_vehicle),
                self.vrp.depots,
            )
        )
        self.actions: sumo_gym.typing.ActionsType = None
        self.rewards: sumo_gym.typing.RewardsType = np.zeros(self.vrp.n_vehicle)

        return {
            "Loading": self.loading,
            "Locations": self.locations,
        }

    def step(self, actions):
        n_vehicle = self.vrp.n_vehicle
        prev_location = self.locations
        self.locations = actions
        fully_loaded = np.asarray([False] * n_vehicle)
        for i in range(n_vehicle):
            capacity_remaining = self.vrp.capacity[i] - self.loading[i]
            location = int(self.locations[i])
            if location in self.vrp.depots:
                fully_loaded[i] = False
                self.loading[i] = 0
            else:
                load_amount = min(self.vrp.demand[location], capacity_remaining)
                fully_loaded[i] = False if capacity_remaining > 0 else True
                self.loading[i] += load_amount
                self.vrp.demand[location] -= load_amount
                self.rewards[i] += (
                    5 * load_amount
                )  # todo: make rewards_rate more customized
                self.rewards[i] -= sumo_gym.utils.calculate_dist(
                    prev_location[i], location, self.vrp.vertices
                )

        self.action_space: sumo_gym.spaces.network.NetworkSpace = (
            sumo_gym.spaces.network.NetworkSpace(
                self.locations,
                self.vrp.get_adj_list(),
                self.vrp.demand,
                fully_loaded,
                self.vrp.depots,
            )
        )

        observation = {
            "Loading": self.loading,
            "Locations": self.locations,
        }
        reward, done = self.rewards, True
        for l in self.locations:
            if l not in self.vrp.depots:
                done = False
                break

        done &= all([True if d == 0 else False for d in self.vrp.demand])
        info = (
            f"Action: {actions}; \nDemand: {self.vrp.demand.astype(int)}"
            + f"; \nReward: {self.rewards.astype(int)}.\n"
        )
        return observation, reward, done, info

    def plot(
        self,
        *,
        ax_dict=None,
        **kwargs: Any,
    ) -> Any:
        import sumo_gym.plot

        return sumo_gym.plot.plot_VRPEnv(self, ax_dict=ax_dict, **kwargs)

    def render(self, mode="human"):
        get_colors = lambda n: list(
            map(lambda i: "#" + "%06x" % random.randint(0x000000, 0x666666), range(n))
        )
        plot_kwargs = {
            "vrp_depot_s": 200,
            "vrp_vertex_s": 200,
            "vrp_depot_c": "darkgreen",
            "vrp_vertex_c": "navy",
            "vrp_depot_marker": r"$\otimes$",
            "vrp_vertex_marker": r"$\odot$",
            "demand_width": 0.4,
            "demand_color": get_colors(self.vrp.n_vertex),
            "loading_width": 0.6,
            "loading_color": get_colors(self.vrp.n_vehicle),
            "location_marker": vehicle_marker,
            "location_s": 2000,
            "location_c": "lightgrey",
        }
        self.plot(**plot_kwargs)



import operator
import random
import sys
import os
import numpy as np
import numpy.typing as npt
from typing import Type, Tuple, Dict, Any
import sumo_gym.typing

import gym
from gym import error, spaces, utils
from gym.utils import seeding
import sumo_gym
from sumo_gym.utils.sumo_utils import SumoRender
from sumo_gym.utils.svg_uitls import vehicle_marker
from sumo_gym.utils.fmp_utils import *


class FMP(object):
    def __init__(
        self,
        mode: str = None,
        net_xml_file_path: str = None,
        demand_xml_file_path: str = None,
        additional_xml_file_path: str = None,
        n_vertex: int = 0,
        n_edge: int = 0,
        n_vehicle: int = 0,
        n_electric_vehicle: int = 0,
        n_charging_station: int = 1,
        vertices: sumo_gym.typing.VerticesType = None,
        demand: sumo_gym.utils.fmp_utils.Demand = None,
        edges: sumo_gym.typing.EdgeType = None,
        electric_vehicles: sumo_gym.utils.fmp_utils.ElectricVehicles = None,
        departures: sumo_gym.typing.DeparturesType = None,
        charging_stations: sumo_gym.utils.fmp_utils.ChargingStation = None,
    ):
        if mode is None:
            raise Exception("Need a mode to identify")
        elif mode == "sumo_config":
            self.__sumo_config_init(
                net_xml_file_path, demand_xml_file_path, additional_xml_file_path
            )
        elif mode == "numerical":
            self.__numerical_init(
                n_vertex,
                n_edge,
                n_vehicle,
                n_electric_vehicle,
                n_charging_station,
                vertices,
                demand,
                edges,
                electric_vehicles,
                departures,
                charging_stations,
            )
        else:
            raise Exception("Need a valid mode")

        if not self._is_valid():
            raise ValueError("FMP setting is not valid")

    def __numerical_init(
        self,
        n_vertex: int = 0,
        n_edge: int = 0,
        n_vehicle: int = 0,
        n_electric_vehicle: int = 0,
        n_charging_station: int = 1,
        vertices: sumo_gym.typing.VerticesType = None,
        demand: sumo_gym.utils.fmp_utils.Demand = None,
        edges: sumo_gym.typing.EdgeType = None,
        electric_vehicles: sumo_gym.utils.fmp_utils.ElectricVehicles = None,
        departures: sumo_gym.typing.DeparturesType = None,
        charging_stations: sumo_gym.utils.fmp_utils.ChargingStation = None,
    ):
        # number
        self.n_vertex = n_vertex
        self.n_edge = n_edge
        self.n_vehicle = n_vehicle
        self.n_electric_vehicle = n_electric_vehicle
        self.n_charging_station = n_charging_station

        # network
        self.vertices = vertices
        self.demand = demand
        self.edges = edges

        # vehicles
        self.electric_vehicles = electric_vehicles
        self.departures = departures
        self.charging_stations = charging_stations

        self.edge_dict = None

    def __sumo_config_init(
        self,
        net_xml_file_path: str = None,
        demand_xml_file_path: str = None,
        additional_xml_file_path: str = None,
    ):
        (
            raw_vertices,  # [id (str), x_coord (float), y_coord (float)]
            raw_charging_stations,  # [id, (x_coord, y_coord), edge_id, charging speed]
            raw_electric_vehicles,  # [id (str), maximum speed (float), maximumBatteryCapacity (float)]
            raw_edges,  # [id (str), from_vertex_id (str), to_vertex_id (str)]
            raw_departures,  # [vehicle_id, starting_edge_id]
            raw_demand,  # [junction_id, dest_vertex_id]
        ) = sumo_gym.utils.xml_utils.decode_xml_fmp(
            net_xml_file_path, demand_xml_file_path, additional_xml_file_path
        )

        # `vertices` is a list of Vertex instances
        # `self.vertex_dict` is a mapping from vertex id in SUMO to idx in vertices
        vertices, self.vertex_dict = convert_raw_vertices(raw_vertices)

        # `edges` is a list of Edge instances
        # `self.edge_dict` is a mapping from SUMO edge id to idx in `edges`
        # `self.edge_length_dict` is a dictionary mapping from SUMO edge id to edge length
        (
            edges,
            self.edge_dict,
            self.edge_length_dict,
        ) = convert_raw_edges(raw_edges, self.vertex_dict)

        # `charging_stations` is a list of ChargingStation instances
        # `self.charging_stations_dict` is a mapping from idx in `charging_stations` to SUMO station id
        (
            charging_stations,
            self.charging_stations_dict,
            self.edge_length_dict,
        ) = convert_raw_charging_stations(
            raw_charging_stations,
            vertices,
            edges,
            self.edge_dict,
            self.edge_length_dict,
        )

        # `electric_vehicles` is a list of ElectricVehicles instances
        # `self.ev_dict` is a mapping from ev sumo id to idx in `electric_vehicles`
        electric_vehicles, self.ev_dict = convert_raw_electric_vehicles(
            raw_electric_vehicles
        )

        # departure should be defined for all vehicles
        # `self.departures` and `self.actual_departures` are lists of indices in `vertices`
        # self.departures[i] is the starting point of electric_vehicles[i] (the endpoint of the passed in edge)
        # self.actual_depatures[i] is the actual start vertex of electric_vehicles[i] (the starting point of the passed in edge)
        departures, actual_departures = convert_raw_departures(
            raw_departures,
            self.ev_dict,
            edges,
            self.edge_dict,
            len(electric_vehicles),
        )

        # `demand` is a list of Demand instances
        demand = convert_raw_demand(raw_demand, self.vertex_dict)

        # set the FMP variables
        self.vertices = np.asarray(vertices)
        self.edges = np.asarray(edges)
        self.charging_stations = np.asarray(charging_stations)
        self.electric_vehicles = np.asarray(electric_vehicles)
        self.departures = np.asarray(departures)
        self.departures = [int(x) for x in self.departures]
        self.actual_departures = np.asarray(actual_departures)
        self.demand = np.asarray(demand)

        self.n_vertex = len(self.vertices)
        self.n_edge = len(self.edges)
        self.n_vehicle = self.n_electric_vehicle = len(self.electric_vehicles)
        self.n_charging_station = len(self.charging_stations)

    def _is_valid(self):
        if (
            not self.n_vertex
            or not self.n_edge
            or not self.n_vehicle
            or not self.n_charging_station
            or self.vertices is None
            or self.charging_stations is None
            or self.electric_vehicles is None
            or self.demand is None
            or self.edges is None
            or self.departures is None
        ):
            return False
        if self.vertices.shape[0] != len(self.vertices):
            return False
        if self.edges.shape[0] != len(self.edges):
            return False
        if self.electric_vehicles.shape[0] != len(self.electric_vehicles):
            return False
        if self.charging_stations.shape[0] != len(self.charging_stations):
            return False
        charging_station_locations = set([cs.location for cs in self.charging_stations])
        for d in self.demand:
            if (
                d.departure in charging_station_locations
                or d.destination in charging_station_locations
            ):
                return False
        # todo: scale judgement
        return True


class FMPEnv(gym.Env):
    metadata = {"render.modes": ["human"]}
    fmp = property(operator.attrgetter("_fmp"))
    __isfrozen = False

    def __init__(self, **kwargs):

        if "mode" not in kwargs:
            raise Exception("Need a mode to identify")
        elif kwargs["mode"] == "sumo_config":
            if "SUMO_GUI_PATH" in os.environ:
                self.sumo_gui_path = os.environ["SUMO_GUI_PATH"]
            else:
                raise Exception("Need 'SUMO_GUI_PATH' in the local environment")

            if "sumo_config_path" in kwargs:
                self.sumo_config_path = kwargs["sumo_config_path"]
                del kwargs["sumo_config_path"]
            else:
                raise Exception("Need 'sumo_config_path' argument to initialize")

            if "render_env" in kwargs:
                self.render_env = kwargs["render_env"]
                del kwargs["render_env"]
            else:
                self.render_env = False

        elif kwargs["mode"] == "numerical":
            if "render_env" in kwargs:
                raise Exception("Only support render for 'sumo_config' mode")

        else:
            raise Exception("Need a valid mode")

        self._fmp = FMP(**kwargs)  # todo: make it "final"
        self.sumo = (
            SumoRender(
                self.sumo_gui_path,
                self.sumo_config_path,
                self.fmp.edge_dict,
                self.fmp.edge_length_dict,
                self.fmp.ev_dict,
                self.fmp.edges,
                self.fmp.n_electric_vehicle,
            )
            if hasattr(self, "render_env") and self.render_env is True
            else None
        )
        self.run = -1
        self._reset()
        self._freeze()

    def __setattr__(self, key, value):
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError(
                "Cannot add new attributes once instance %r is initialized" % self
            )
        object.__setattr__(self, key, value)

    def _freeze(self):
        self.__isfrozen = True

    def reset(self):
        return self._reset()

    def _reset(self):
        self.states = [FMPState() for _ in range(self.fmp.n_electric_vehicle)]
        self.responded = set()
        for i in range(self.fmp.n_electric_vehicle):
            self.states[i].location = self.fmp.departures[i]
            self.states[i].battery = self.fmp.electric_vehicles[i].capacity

        self.action_space: sumo_gym.spaces.grid.GridSpace = (
            sumo_gym.spaces.grid.GridSpace(
                self.fmp.vertices,
                self.fmp.demand,
                self.responded,
                self.fmp.edges,
                self.fmp.electric_vehicles,
                self.fmp.charging_stations,
                self.states,
                self.sumo if hasattr(self, "sumo") else None,
            )
        )
        self.actions: sumo_gym.typing.ActionsType = None
        self.rewards: sumo_gym.typing.RewardsType = np.zeros(self.fmp.n_vehicle)

        observation = {
            "Locations": [s.location for s in self.states],
            "Batteries": [s.battery for s in self.states],
            "Is_loading": [s.is_loading for s in self.states],
            "Is_charging": [s.is_charging for s in self.states],
            "Takes_action": [
                True
                if s.is_loading.target == NO_LOADING
                and s.is_charging.target == NO_CHARGING
                else False
                for s in self.states
            ],
        }
        return observation

    def step(self, actions):
        prev_locations = []
        travel_info = []
        self.rewards = np.zeros(self.fmp.n_vehicle)
        for i in range(self.fmp.n_vehicle):
            prev_location = self.states[i].location
            prev_locations.append(prev_location)
            prev_is_loading = self.states[i].is_loading.current
            prev_battery = self.states[i].battery

            self.states[i].battery -= sumo_gym.utils.fmp_utils.dist_between(
                self.fmp.vertices,
                self.fmp.edges,
                prev_location
                if actions[i].location == IDLE_LOCATION
                else actions[i].location,
                prev_location,
            )

            (
                self.states[i].is_loading,
                self.states[i].is_charging,
                self.states[i].location,
            ) = (
                Loading(actions[i].is_loading.current, actions[i].is_loading.target),
                Charging(actions[i].is_charging.current, actions[i].is_charging.target),
                actions[i].location,
            )
            travel_info.append((prev_location, actions[i].location))

            # assert self.states[i].battery >= 0
            if self.states[i].is_charging.current != NO_CHARGING:
                self.states[i].battery += self.fmp.charging_stations[
                    self.states[i].is_charging.current
                ].charging_speed

            if prev_is_loading != -1 and self.states[i].is_loading.current == -1:
                self.responded.add(prev_is_loading)
                self.rewards[i] += sumo_gym.utils.fmp_utils.get_hot_spot_weight(
                    self.fmp.vertices,
                    self.fmp.edges,
                    self.fmp.demand,
                    self.fmp.demand[prev_is_loading].departure,
                ) * sumo_gym.utils.fmp_utils.dist_between(
                    self.fmp.vertices,
                    self.fmp.edges,
                    self.fmp.demand[prev_is_loading].departure,
                    self.fmp.demand[prev_is_loading].destination,
                )

        print("Batteries:", [s.battery for s in self.states])
        print("Rewards:", self.rewards)
        observation = {
            "Locations": [s.location for s in self.states],
            "Batteries": [s.battery for s in self.states],
            "Is_loading": [s.is_loading for s in self.states],
            "Is_charging": [s.is_charging for s in self.states],
            "Takes_action": [
                True
                if s.is_loading.target == NO_LOADING
                and s.is_charging.target == NO_CHARGING
                and not s.location == IDLE_LOCATION
                else False
                for s in self.states
            ],
        }
        reward, done, info = (
            self.rewards,
            self.responded == set(range(len(self.fmp.demand))),
            "",
        )

        if hasattr(self, "sumo") and self.sumo is not None:
            self.sumo.update_travel_vertex_info_for_vehicle(travel_info)
            self.render()

        return observation, reward, done, info

    def plot(
        self,
        *,
        ax_dict=None,
        **kwargs: Any,
    ) -> Any:
        get_colors = lambda n: list(
            map(lambda i: "#" + "%06x" % random.randint(0x000000, 0x666666), range(n))
        )
        plot_kwargs = {
            "fmp_vertex_s": 200,
            "fmp_vertex_c": "navy",
            "fmp_vertex_marker": r"$\odot$",
            "demand_width": 0.4,
            "demand_color": get_colors(self.fmp.n_vertex),
            "loading_width": 0.6,
            "loading_color": get_colors(self.fmp.n_vehicle),
            "location_marker": vehicle_marker,
            "location_s": 2000,
            "location_c": "lightgrey",
        }
        self.plot(**plot_kwargs)

    def render(self, mode="human"):
        # todo: this part should be move to .render()
        if self.sumo_gui_path is None:
            raise EnvironmentError("Need sumo-gui path to render")
        elif self.sumo is not None:
            self.sumo.render()

    # TODO: need to add default behavior also
    def close(self):
        if hasattr(self, "sumo") and self.sumo is not None:
            self.sumo.close()



from typing import Type, Tuple, Any, List
import numpy as np
import numpy.typing as npt
from gym.spaces import Space
import sumo_gym


VertexType = Tuple[float]
VerticesType = npt.NDArray[VertexType]
DemandType = npt.NDArray[float]
EdgeType = npt.NDArray[Tuple[int]]
DeparturesType = npt.NDArray[int]
CapacityType = npt.NDArray[float]
AdjListType = npt.NDArray[npt.NDArray[int]]
LocationsType = npt.NDArray[int]
LoadingType = npt.NDArray[Tuple[float]]
SpaceType = Space
ActionsType = npt.NDArray[int]
RewardsType = npt.NDArray[float]



import numpy as np
import numpy.typing as npt
import gym
from typing import Tuple

import sumo_gym.typing


class NetworkSpace(gym.spaces.Space):
    def __init__(
        self,
        locations: sumo_gym.typing.LocationsType,
        adj_list: sumo_gym.typing.AdjListType,
        demand: sumo_gym.typing.DemandType,
        fully_loaded: npt.NDArray[bool],
        depots: npt.NDArray[int],
        shape=None,
        dtype=None,
        seed=None,
    ):
        super(NetworkSpace, self).__init__()
        self.locations = locations
        self.adj_list = adj_list
        self.demand = demand
        self.fully_loaded = fully_loaded
        self.depots = depots

    def sample(self) -> npt.NDArray[int]:
        samples = np.copy(self.locations)
        for i, loc in enumerate(self.locations):
            if self.fully_loaded[i]:
                samples[i] = np.random.choice(self.depots)
            else:
                # todo: need to bypass to search the demand
                # currently, we are not bypassing, i.e., never go to the vertices that are not connected
                destination = []
                for v in self.adj_list[loc]:
                    if self.demand[v] or v in self.depots:  # if is depot or have demand
                        destination.append(v)

                if len(destination):
                    samples[i] = np.random.choice(destination, 1)[0]

        # avoid crash
        for j, s in enumerate(samples):
            if s in np.delete(samples, j) and s not in self.depots:
                samples[j] = self.locations[j]

        return samples.astype(int)



from gym.envs.registration import register
from sumo_gym.envs.vrp import VRP, VRPEnv
from sumo_gym.spaces.network import NetworkSpace
from sumo_gym.utils.xml_utils import decode_xml
from sumo_gym.utils.network_utils import (
    calculate_dist,
    get_adj_from_list,
    get_adj_to_list,
)
from sumo_gym.utils.svg_uitls import vehicle_marker
from typing import Tuple


__all__ = (
    "VRP",
    "VRPEnv",
    "NetworkSpace",
    "decode_xml",
    "calculate_dist",
    "get_adj_from_list",
    "get_adj_to_list",
    "vehicle_marker",
)


def __dir__() -> Tuple[str, ...]:
    return __all__


register(
    id="VRP-v0",
    entry_point="sumo_gym.envs:VRPEnv",
)

register(
    id="FMP-v0",
    entry_point="sumo_gym.envs:FMPEnv",
)



from typing import Any
import random
from random import randrange

import sumo_gym
from sumo_gym.utils.fmp_utils import (
    IDLE_LOCATION,
    Loading,
    Charging,
    GridAction,
    NO_LOADING,
    NO_CHARGING,
)
import gym

import numpy as np
import numpy.typing as npt

from sumo_gym.utils.sumo_utils import SumoRender


class GridSpace(gym.spaces.Space):
    def __init__(
        self,
        vertices: sumo_gym.typing.VerticesType = None,
        demand: sumo_gym.utils.fmp_utils.Demand = None,
        responded: set = None,
        edges: sumo_gym.typing.EdgeType = None,
        electric_vehicles: sumo_gym.utils.fmp_utils.ElectricVehicles = None,
        charging_stations: sumo_gym.utils.fmp_utils.ChargingStation = None,
        states=None,
        sumo: SumoRender = None,
        shape=None,
        dtype=None,
        seed=None,
    ):
        super(GridSpace, self).__init__()
        self.vertices = vertices
        self.demand = demand
        self.responded = responded
        self.edges = edges
        self.electric_vehicles = electric_vehicles
        self.charging_stations = charging_stations
        self.states = states
        self.sumo = sumo

    def sample(
        self,
    ) -> Any:
        n_vehicle = len(self.states)
        samples = [GridAction(state) for state in self.states]

        # extract all demands being responded
        responding = set()
        for i in range(n_vehicle):
            if self.states[i].is_loading.current != NO_LOADING:
                responding.add(self.states[i].is_loading.current)
            elif self.states[i].is_loading.target != NO_LOADING:
                responding.add(self.states[i].is_loading.target)  # todo

        if self.sumo is None:
            stop_statuses = [True] * n_vehicle
        else:
            stop_statuses = self.sumo.get_stop_status()

        for i in range(n_vehicle):
            if (
                self.states[i].location == IDLE_LOCATION
            ):  # idle vehicle remove from simulation, skip action
                print("----- IDLE...")
                samples[i] = self.states[i]
                continue

            if not stop_statuses[i]:
                print("----- Traveling along the edge...")
                continue

            if self.states[i].is_loading.current != NO_LOADING:  # is on the way
                print("----- In the way of demand:", self.states[i].is_loading.current)
                loc = sumo_gym.utils.fmp_utils.one_step_to_destination(
                    self.vertices,
                    self.edges,
                    self.states[i].location,
                    self.demand[self.states[i].is_loading.current].destination,
                )

                if (
                    self.states[i].location
                    == self.demand[self.states[i].is_loading.current].destination
                ):
                    samples[i].is_loading = Loading(NO_LOADING, NO_LOADING)
                else:
                    samples[i].is_loading = Loading(
                        self.states[i].is_loading.current,
                        self.states[i].is_loading.target,
                    )
                samples[i].location = loc
            elif self.states[i].is_loading.target != NO_LOADING:  # is to the way
                print("----- In the way to respond:", self.states[i].is_loading.target)
                loc = sumo_gym.utils.fmp_utils.one_step_to_destination(
                    self.vertices,
                    self.edges,
                    self.states[i].location,
                    self.demand[self.states[i].is_loading.target].departure,
                )
                samples[i].location = loc
                if loc == self.demand[self.states[i].is_loading.target].departure:
                    samples[i].is_loading = Loading(
                        self.states[i].is_loading.target,
                        self.states[i].is_loading.target,
                    )
                else:
                    samples[i].is_loading = Loading(
                        self.states[i].is_loading.current,
                        self.states[i].is_loading.target,
                    )
            elif self.states[i].is_charging.current != NO_CHARGING:  # is charging
                samples[i].location = self.charging_stations[
                    self.states[i].is_charging.current
                ].location
                if (
                    self.electric_vehicles[i].capacity - self.states[i].battery
                    > self.charging_stations[
                        self.states[i].is_charging.current
                    ].charging_speed
                ):
                    print("----- Still charging")
                    samples[i].is_charging = self.states[i].is_charging
                else:
                    print("----- Charging finished")
                    samples[i].is_charging = Charging(NO_CHARGING, NO_CHARGING)
            elif (
                self.states[i].is_charging.target != NO_CHARGING
            ):  # is on the way to charge
                if (
                    self.states[i].location
                    == self.charging_stations[
                        self.states[i].is_charging.target
                    ].location
                ):
                    print(
                        "----- Arrived charging station:",
                        self.states[i].is_charging.target,
                    )
                    samples[i].is_charging = Charging(
                        self.states[i].is_charging.target,
                        self.states[i].is_charging.target,
                    )
                else:
                    print(
                        "----- In the way to charge:", self.states[i].is_charging.target
                    )
                    loc = sumo_gym.utils.fmp_utils.one_step_to_destination(
                        self.vertices,
                        self.edges,
                        self.states[i].location,
                        self.charging_stations[
                            self.states[i].is_charging.target
                        ].location,
                    )
                    samples[i].location = loc
                    samples[i].is_charging = Charging(
                        self.states[i].is_charging.current,
                        self.states[i].is_charging.target,
                    )
            else:  # available
                diagonal_len = 2 * (
                    float(max(self.vertices, key=lambda item: item.y).y)
                    - float(min(self.vertices, key=lambda item: item.y).y)
                    + float(max(self.vertices, key=lambda item: item.x).x)
                    - float(min(self.vertices, key=lambda item: item.x).x)
                )
                probability_of_togo_charge = self.states[i].battery / (
                    diagonal_len - self.electric_vehicles[i].capacity
                ) + self.electric_vehicles[i].capacity / (
                    self.electric_vehicles[i].capacity - diagonal_len
                )
                if np.random.random() < probability_of_togo_charge:
                    rcs = randrange(len(self.charging_stations))
                    print("----- Goto charge:", rcs)
                    loc = sumo_gym.utils.fmp_utils.one_step_to_destination(
                        self.vertices,
                        self.edges,
                        self.states[i].location,
                        self.charging_stations[rcs].location,
                    )
                    samples[i].location = loc
                    samples[i].is_charging.target = rcs
                else:
                    available_dmd = [
                        d
                        for d in range(len(self.demand))
                        if d not in self.responded and d not in responding
                    ]
                    if len(available_dmd):
                        dmd_idx = random.choices(available_dmd)[0]
                        print("----- Choose dmd_idx:", dmd_idx)
                        responding.add(dmd_idx)
                        loc = sumo_gym.utils.fmp_utils.one_step_to_destination(
                            self.vertices,
                            self.edges,
                            self.states[i].location,
                            self.demand[dmd_idx].departure,
                        )
                        samples[i].location = loc
                        if loc == self.demand[dmd_idx].departure:
                            samples[i].is_loading = Loading(dmd_idx, dmd_idx)
                        else:
                            samples[i].is_loading = Loading(NO_LOADING, dmd_idx)
                    else:
                        print("----- IDLE...")
                        samples[i].location = IDLE_LOCATION

        print("Samples: ", samples)
        return samples



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

extras_require = {
    "plot": [
        "matplotlib >=3.0",
    ],
    "test": [
        "pytest >=6",
        "pytest-mpl >=0.12",
    ],
    "torch": [
        "torch >=1.10.0",
        "torchvision >=0.11.1",
        "torchaudio >=0.10.0",
    ],
}

extras_require["dev"] = [
    *extras_require["test"],
    *extras_require["plot"],
    *extras_require["torch"],
    "ipykernel",
]

extras_require["docs"] = [
    *extras_require["plot"],
    "nbsphinx",
    "Sphinx >=3.0.0",
    "sphinx_copybutton",
    "sphinx_rtd_theme >=0.5.0",
    "sphinx_book_theme >=0.0.38",
    "ipython",
    "ipykernel",
]

extras_require["all"] = sorted(set(sum(extras_require.values(), [])))

setuptools.setup(
    name="sumo_gym",
    version="0.3.0",
    author="Shuo Liu",
    author_email="ninomyemail@gmail.com",
    description="OpenAI-gym like toolkit for developing and comparing reinforcement learning algorithms on SUMO",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lovelybuggies/sumo-gym",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7",
    install_requires=["gym"],
    extras_require=extras_require,
)
